@@ -1,7 +1,6 @@
 import asyncio
 import lusid
 import pandas as pd
-import numpy as np
 import json
 from lusidtools import cocoon
 from lusidtools.cocoon.async_tools import run_in_executor, ThreadPool
@@ -10,11 +9,12 @@ from lusidtools.cocoon.utilities import (
     checkargs,
     strip_whitespace,
     group_request_into_one,
+    extract_unique_portfolio_codes,
+    extract_unique_portfolio_codes_effective_at_tuples,
 )
 from lusidtools.cocoon.validator import Validator
 from datetime import datetime
 import pytz
-from lusidtools.logger import LusidLogger
 import logging
 
 
@@ -713,6 +713,7 @@ async def _construct_batches(
     domain_lookup: dict,
     sub_holding_keys: list,
     sub_holding_keys_scope: str,
+    return_unmatched_items: bool,
     **kwargs,
 ):
     """
@@ -744,6 +745,8 @@ async def _construct_batches(
         The sub holding keys to use
     sub_holding_keys_scope : str
         The scope to use for the sub-holding keys
+    return_unmatched_items : bool
+        Whether items with unmatched identifiers should be returned for transaction or holding upserts
     kwargs
         Arguments specific to each call e.g. effective_at for holdings
 
@@ -901,11 +904,322 @@ async def _construct_batches(
             raise response
 
     # Collects the exceptions as failures and successful calls as values
-    return {
+    returned_response = {
         "errors": [r for r in responses_flattened if isinstance(r, Exception)],
         "success": [r for r in responses_flattened if not isinstance(r, Exception)],
     }
 
+    # For successful transactions or holdings file types, optionally return unmatched identifiers with the responses
+    if check_for_unmatched_items(flag=return_unmatched_items, file_type=file_type,):
+        returned_response["unmatched_items"] = unmatched_items(
+            api_factory=api_factory,
+            scope=kwargs.get("scope", None),
+            data_frame=data_frame,
+            mapping_required=mapping_required,
+            file_type=file_type,
+            returned_response=returned_response,
+            sync_batches=sync_batches,
+        )
+
+    return returned_response
+
+
+def check_for_unmatched_items(flag, file_type):
+    """
+    This method contains the conditional logic to determine whether the unmatched_items validation should be run.
+    It should not be run if:
+        a) it was not requested
+        b) the upload was not for transactions or holdings
+
+    Parameters
+    ----------
+    flag : bool
+        The bool flag that indicates whether any unmatched identifiers should be checked
+    file_type : str
+        The file type of the upload
+    Returns
+    -------
+
+    A boolean value indicating whether the unmatched_items validation should be completed
+    """
+    condition_1 = flag is True
+    condition_2 = file_type in ["transaction", "holding"]
+
+    return condition_1 and condition_2
+
+
+@checkargs
+def unmatched_items(
+    api_factory: lusid.utilities.ApiClientFactory,
+    scope: str,
+    data_frame: pd.DataFrame,
+    mapping_required: dict,
+    file_type: str,
+    returned_response: dict,
+    sync_batches: list = None,
+):
+    """
+    This method orchestrates the identification of holdings or transactions objects that were successfully uploaded
+    but did not resolve to known identifiers (i.e. where the instrument_uid is LUID_ZZZZZZ).
+
+    If there were any errors in the attempted upload, the check will not complete and an error message will be returned
+    to let the user know that errors will need to be resolved first.
+
+    Parameters
+    ----------
+    api_factory : lusid.utilities.ApiClientFactory api_factory
+        The api factory to use
+    scope : str
+        The scope of the resource to load the data into
+    data_frame : pd.DataFrame
+        The DataFrame containing the data
+    mapping_required : dict
+        The required mapping
+    file_type : str
+        The type of file e.g. transactions, instruments, holdings, quotes, portfolios
+    returned_response : dict
+        The response from laod_from_data_frame
+    sync_batches : list
+        A list of the batches used to upload the data into LUSID.
+
+    Returns
+    -------
+    responses: list
+        A list of objects to be appended to the ultimate response for load_from_data_frame.
+    """
+
+    if len(returned_response["errors"]) > 0:
+        return ["Please resolve all upload errors to check for unmatched items."]
+
+    if file_type == "transaction":
+        return _unmatched_transactions(
+            api_factory=api_factory,
+            scope=scope,
+            data_frame=data_frame,
+            mapping_required=mapping_required,
+            sync_batches=sync_batches,
+        )
+    elif file_type == "holding":
+        return _unmatched_holdings(
+            api_factory=api_factory, scope=scope, sync_batches=sync_batches,
+        )
+
+
+def _unmatched_transactions(
+    api_factory: lusid.utilities.ApiClientFactory,
+    scope: str,
+    data_frame: pd.DataFrame,
+    mapping_required: dict,
+    sync_batches: list = None,
+):
+    """
+    This method identifies which instruments were not resolved with a transaction upload using load_from_data_frame.
+
+    Parameters
+    ----------
+    api_factory : lusid.utilities.ApiClientFactory api_factory
+        The api factory to use
+    scope : str
+        The scope of the resource to load the data into
+    data_frame : pd.DataFrame
+        The DataFrame containing the data
+    sync_batches : list
+        A list of the batches used to upload the data into LUSID.
+
+    Returns
+    -------
+    responses: list
+        A list of transaction objects to be appended to the ultimate response for load_from_data_frame.
+    """
+    # Extract a list of portfolio codes from the sync_batches
+    portfolio_codes = extract_unique_portfolio_codes(sync_batches)
+
+    # Create empty list to hold transaction ids and instruments
+    unmatched_transactions = []
+
+    # For each portfolio, request the unmatched transactions from LUSID and append to the instantiated list
+    for portfolio_code in portfolio_codes:
+        unmatched_transactions.extend(
+            return_unmatched_transactions(
+                api_factory=api_factory, scope=scope, code=portfolio_code,
+            )
+        )
+
+    # With the upload dataframe, filter out any transactions from the list that were not part of this upload and return
+    return filter_unmatched_transactions(
+        data_frame=data_frame,
+        mapping_required=mapping_required,
+        unmatched_transactions=unmatched_transactions,
+    )
+
+
+def return_unmatched_transactions(
+    api_factory: lusid.utilities.ApiClientFactory, scope: str, code: str
+):
+    """
+    Call the get transactions api and only return those transactions with unresolved identifiers.
+
+    Multiple pages of responses for the get_transactions api call are handled.
+
+    Parameters
+    ----------
+    api_factory : lusid.utilities.ApiClientFactory api_factory
+        The api factory to use
+    scope : str
+        The scope of the resource to load the data into
+    code : str
+        The code of the portfolio containing the transactions we want to return
+
+    Returns
+    -------
+    A list of transaction objects with the structure.
+    """
+    transactions_api = api_factory.build(lusid.api.TransactionPortfoliosApi)
+    done = False
+    next_page = None
+    unmatched_transactions = []
+
+    # We need to handle the possibility of paginated results
+    while not done:
+
+        # There must be a filter included so only transactions with unmatched instruments are returned
+        kwargs = {
+            "scope": scope,
+            "code": code,
+            "filter": "instrumentUid eq'LUID_ZZZZZZZZ'",
+        }
+
+        if next_page is not None:
+            kwargs["page"] = next_page
+
+        response = transactions_api.get_transactions(**kwargs)
+
+        unmatched_transactions.extend([response for response in response.values])
+
+        next_page = response.next_page
+        done = response.next_page is None
+
+    return unmatched_transactions
+
+
+def filter_unmatched_transactions(
+    data_frame: pd.DataFrame, mapping_required: dict, unmatched_transactions: list,
+):
+    """
+    This method will take the full list of unmatched transactions and remove any transactions that were not
+    part of the current upload with load_from_data_frame.
+
+    Parameters
+    ----------
+    data_frame : pd.DataFrame
+        The DataFrame containing the data
+    mapping_required : dict
+        The required mapping from load_from_data_frame that helps identify the transaction_id column in the DataFrame
+    unmatched_transactions : list
+         A list of transaction objects.
+
+    Returns
+    -------
+    A filtered list of transaction objects.
+    """
+    # Create a unique list of transaction_ids from the dataframe
+    valid_txids = set(data_frame[mapping_required.get("transaction_id")])
+
+    # Iterate through the transactions and if the transaction_id is in the set, add it to the list to be returned
+    filtered_unmatched_transactions = [
+        unmatched_transaction
+        for unmatched_transaction in unmatched_transactions
+        if unmatched_transaction.transaction_id in valid_txids
+    ]
+
+    return filtered_unmatched_transactions
+
+
+def _unmatched_holdings(
+    api_factory: lusid.utilities.ApiClientFactory,
+    scope: str,
+    sync_batches: list = None,
+):
+    """
+    This method identifies which instruments were not resolved with a holdings upload using load_from_data_frame.
+
+    Parameters
+    ----------
+    api_factory : lusid.utilities.ApiClientFactory api_factory
+        The api factory to use
+    scope : str
+        The scope of the resource to load the data into
+    sync_batches : list
+        A list of the batches used to upload the data into LUSID
+
+    Returns
+    -------
+    responses: list
+        A list of holding objects to be appended to the ultimate response for load_from_data_frame.
+    """
+    # Extract a list of tuples of portfolio codes and effective at times from sync_batches
+    code_tuples = extract_unique_portfolio_codes_effective_at_tuples(sync_batches)
+
+    # Create empty list to hold holdings that have not been resolved
+    unmatched_holdings = []
+
+    # For each holding adjustment in the upload, check whether any contained unresolved instruments and append to list
+    for code_tuple in code_tuples:
+        unmatched_holdings.extend(
+            return_unmatched_holdings(
+                api_factory=api_factory, scope=scope, code_tuple=code_tuple,
+            )
+        )
+
+    return unmatched_holdings
+
+
+def return_unmatched_holdings(
+    api_factory: lusid.utilities.ApiClientFactory, scope: str, code_tuple: (str, str),
+):
+    """
+    Call the get holdings adjustments api and return a list of holding objects that have unresolved identifiers.
+
+    Parameters
+    ----------
+    api_factory : lusid.utilities.ApiClientFactory api_factory
+        The api factory to use
+    scope : str
+        The scope of the resource to load the data into
+    code_tuple : (str, str)
+        A tuple of format (portfolio_code, effective_at) which represents the combination of values
+        used when upserting holdings during load_from_data_frame for file_type 'holding'
+
+    Returns
+    -------
+    A list of holding objects.
+
+    """
+    transactions_api = api_factory.build(lusid.api.TransactionPortfoliosApi)
+
+    # In case a holdings check occurs for a portfolio code and effective at combination that contain no holdings,
+    # make sure to gracefully handle the exception so that any LUSID error from the main upload is not suppressed.
+    try:
+        response = transactions_api.get_holdings_adjustment(
+            scope=scope,
+            code=code_tuple[0],
+            effective_at=str(DateOrCutLabel(code_tuple[1])),
+        )
+    except lusid.ApiException as e:
+        if "HoldingsAdjustmentDoesNotExist" in str(e.body):
+            logging.info(
+                f"While validating holding upload, LUSID was unable to find any holdings within portfolio "
+                f"{code_tuple[0]} for effectiveAt {str(DateOrCutLabel(code_tuple[1]))}."
+            )
+            return []
+
+    # Create a list of holding objects with unmatched identifiers (e.g. with LUID_ZZZZZZZZ) from the response
+    return [
+        adjustment
+        for adjustment in response.adjustments
+        if adjustment.instrument_uid == "LUID_ZZZZZZZZ"
+    ]
+
 
 @checkargs
 def load_from_data_frame(
@@ -925,6 +1239,7 @@ def load_from_data_frame(
     holdings_adjustment_only: bool = False,
     thread_pool_max_workers: int = 5,
     sub_holding_keys_scope: str = None,
+    return_unmatched_items: bool = False,
 ):
     """
 
@@ -963,6 +1278,11 @@ def load_from_data_frame(
         The maximum number of workers to use in the thread pool used by the function
     sub_holding_keys_scope : str
         The scope to add the sub-holding keys to
+    return_unmatched_items : bool
+        When loading transactions or holdings, a 'True' flag will return a list of the transaction or holding
+        objects where their instruments were unmatched at the time of the upsert
+        (i.e. where instrument_uid is LUID_ZZZZZZ). This parameter will be ignored for file types other than
+        transactions or holdings
     Returns
     -------
     responses: dict
@@ -1303,6 +1623,7 @@ def load_from_data_frame(
             domain_lookup=domain_lookup,
             sub_holding_keys=sub_holding_keys,
             sub_holding_keys_scope=sub_holding_keys_scope,
+            return_unmatched_items=return_unmatched_items,
             **keyword_arguments,
         ),
         loop,
